#include <jni.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include "UnixSocket.h"

/* -------------------- shared -------------------- */

#define FN_SOCKET 1
#define FN_BIND 2
#define FN_LISTEN 3
#define FN_CONNECT 4
#define FN_ACCEPT 5
#define FN_CLOSE 6
#define FN_GETBYTEARRAYELEMENTS 7
#define FN_RECV 8
#define FN_SEND 9

#define THROW_FLAG ((jlong)1 << 63)
#define OK (jlong)0

jlong throw(int function, int error) {
	/* printf("throwing: %i %i\n", function, error) ; */
	jlong lf = function ;
	jlong le = error ;
	return THROW_FLAG | ((lf & 0xff) << 32) | (le & 0xffffffff) ;
}

jlong ret(int value) {
	/* printf("returning: %i\n", value) ; */
	jlong lv = value ;
	return (lv & 0xffffffff) ;
}

size_t sunsize(struct sockaddr_un *sun) {
	return sizeof((*sun).sun_family) + strlen((*sun).sun_path) + 1 ;
}

/* -------------------- UnixSocket -------------------- */

/*
 * Class:     UnixSocket
 * Method:    _open
 * Signature: (Ljava/lang/String;)J
 */
JNIEXPORT jlong JNICALL Java_UnixSocket__1open(JNIEnv *env, jclass class, jstring addr) {
	int sock = socket(AF_UNIX, SOCK_STREAM, 0) ;
	if (sock == -1) {
		return throw(FN_SOCKET, errno) ;
	}
	const char *addrStr = (*env)->GetStringUTFChars(env, addr, NULL) ;
	if (addrStr == NULL) {
		close(sock) ;
		return 0 ; /* OutOfMemoryError already thrown by JNI */
	}
	struct sockaddr_un addrStruct ;
	memset(&addrStruct, 0, sizeof(addrStruct)) ;
	addrStruct.sun_family = AF_UNIX ;
	strcpy(addrStruct.sun_path, addrStr) ;
	(*env)->ReleaseStringUTFChars(env, addr, addrStr) ;
	int ok = connect(sock, (struct sockaddr *)&addrStruct, sunsize(&addrStruct)) ;
	if (ok != 0) {
		close(sock) ;
		return throw(FN_CONNECT, errno) ;
	}
	return ret(sock) ;
}

/*
 * Class:     UnixSocket
 * Method:    _read
 * Signature: (I[BII)J
 */
JNIEXPORT jlong JNICALL Java_UnixSocket__1read(JNIEnv *env, jclass class, jint sock, jbyteArray buf, jint off, jint len) {
	signed char *cbuf = (*env)->GetByteArrayElements(env, buf, NULL) ;
	if (cbuf == NULL) {
		return throw(FN_GETBYTEARRAYELEMENTS, 0) ;
	}
	ssize_t n = recv(sock, (cbuf + off), len, 0) ;
	(*env)->ReleaseByteArrayElements(env, buf, cbuf, 0) ;
	if (n < 0) {
		return throw(FN_RECV, errno) ;
	}
	return ret(n) ;
}

/*
 * Class:     UnixSocket
 * Method:    _write
 * Signature: (I[BII)J
 */
JNIEXPORT jlong JNICALL Java_UnixSocket__1write(JNIEnv *env, jclass class, jint sock, jbyteArray buf, jint off, jint len) {
	signed char *cbuf = (*env)->GetByteArrayElements(env, buf, NULL) ;
	if (cbuf == NULL) {
		return throw(FN_GETBYTEARRAYELEMENTS, 0) ;
	}
	ssize_t n = send(sock, (cbuf + off), len, 0) ;
	(*env)->ReleaseByteArrayElements(env, buf, cbuf, JNI_ABORT) ;
	if (n < 0) {
		return throw(FN_SEND, errno) ;
	}
	return ret(n) ;
}

/*
 * Class:     UnixSocket
 * Method:    _close
 * Signature: (I)J
 */
JNIEXPORT jlong JNICALL Java_UnixSocket__1close(JNIEnv *env, jclass class, jint sock) {
	int ok = close(sock) ;
	if (ok != 0) {
		return throw(FN_CLOSE, errno) ;
	}
	return OK ;
}

/* -------------------- UnixServerSocket -------------------- */

     int
     unlink(const char *path);

/*
 * Class:     UnixServerSocket
 * Method:    _open
 * Signature: (Ljava/lang/String;)J
 */
JNIEXPORT jlong JNICALL Java_UnixServerSocket__1open(JNIEnv *env, jclass class, jstring addr, jint backlog) {
	int sock = socket(AF_UNIX, SOCK_STREAM, 0) ;
	if (sock == -1) {
		return throw(FN_SOCKET, errno) ;
	}
	const char *addrStr = (*env)->GetStringUTFChars(env, addr, NULL) ;
	if (addrStr == NULL) {
		close(sock) ;
		return 0 ; /* OutOfMemoryError already thrown by JNI */
	}
	struct sockaddr_un addrStruct ;
	memset(&addrStruct, 0, sizeof(addrStruct)) ;
	addrStruct.sun_family = AF_UNIX ;
	strcpy(addrStruct.sun_path, addrStr) ;
	(*env)->ReleaseStringUTFChars(env, addr, addrStr) ;
	int ok ;
	ok = bind(sock, (struct sockaddr *)&addrStruct, sunsize(&addrStruct)) ;
	if (ok != 0) {
		close(sock) ;
		return throw(FN_BIND, errno) ;
	}
	ok = listen(sock, backlog) ;
	if (ok != 0) {
		close(sock) ;
		return throw(FN_LISTEN, errno) ;
	}
	return ret(sock) ;
}

/*
 * Class:     UnixServerSocket
 * Method:    _accept
 * Signature: (I)J
 */
JNIEXPORT jlong JNICALL Java_UnixServerSocket__1accept(JNIEnv *env, jclass class, jint sock) {
	int asock = accept(sock, NULL, 0) ;
	if (asock == -1) {
		return throw(FN_ACCEPT, errno) ;
	}
	return ret(asock) ;
}

/*
 * Class:     UnixServerSocket
 * Method:    _close
 * Signature: (I)J
 */
JNIEXPORT jlong JNICALL Java_UnixServerSocket__1close (JNIEnv *env, jclass class, jint sock) {
	int ok = close(sock) ;
	if (ok != 0) {
		return throw(FN_CLOSE, errno) ;
	}
	return OK ;
}
